import os, sys
if 'nt' in sys.builtin_module_names:
    import winapps

from typing import Dict, List, Optional, Tuple
from theng.data import JSONFile, DataFile
from theng.data.pathfinder import Trigger, TriggerCollection, Door, DoorCollection,\
    Group, GroupCollection, MeasurementRegion, MeasurementRegionCollection,\
    Occupant, OccupantCollection, Room, RoomCollection, Summary, Target, TargetCollection
from theng.models import Model
from theng.args import get_arguments

"""This Dict defines all of the known results files.
Each item should have a descriptive key, used later by the user to access the files,
and a list of tuples representing the files associated with the key.
The first item in the tuple is the file type. This will be nested under the descriptive
key. The second element of the tuple is the filename format.
This format accepts wildcard "*" characters conforming to the glob package's usage, with recurse enabled.
This format will be prepended with the model base name for parsing
"""
_known_output_files: Dict[str, List[Tuple]] = {
    "triggers": [
        ("csv", "_triggers.csv"), 
        ("json", "_triggers.json")
    ],
    "doors": [
        ("csv", "_doors.csv"),
        ("json", "_doors.json")
    ],
    "groups": [
        ("csv", "_groups.csv"),
        ("json", "_groups.json")
    ],
    "measurement_regions": [
        ("csv", "_measurement-regions.csv"),
        ("json", "_measurement-regions.json")
    ],
    "targets": [
        ("csv", "_occtargets.csv"),
        ("json", "_occtargets.json")
    ],
    "occupants_detailed": [
        ("csv", "_occupants_detailed.csv"),
        ("json", "_occupants_detailed.json"),
        ("csv", "_occupant_*_*.csv"),
        ("json", "_occupant_*_*.json")
    ],
    "occupant_params": [
        ("csv", "_occupant_params.csv"),
        ("json", "_occupant_params.json")
    ],
    "occupants": [
        ("csv", "_occupants.csv"),
        ("json", "_occupants.json")
    ],
    "cummulative": [
        ("json", "_out.json")
    ],
    "rooms": [
        ("csv", "_rooms.csv"),
        ("json", "_rooms.json")
    ],
    "social_distance_accumulated": [
        ("csv", "_sd_accumulated.csv"),
        ("json", "_sd_accumulated.json")
    ],
    "social_distance_transient": [
        ("csv", "_sd_transient.csv"),
        ("json", "_sd_transient.json")
    ],
    "summary": [
        ("txt", "_summary.txt"),
        ("json", "_summary.json")
    ],
    "views": [
        ("json", "_views.json")
    ],
    "octrmesh": [
        ("octrmesh", ".octrmesh")
    ],
    "pfd": [
        ("pfd", ".pfd")
    ],
    "pfg": [
        ("pfg", ".pfg")
    ],
    "input": [
        ("txt", ".txt")
    ],
    "performance": [
        ("txt", "_performance.txt")
    ]
}


class PathfinderModel(Model):
    """Represents a single Pathfinder simulation and it's results

    Args:
        results_file_name (str): The basename of the .pfr Results file generated by Pathfinder
    
    Attributes: 
        triggers (TriggerCollection): Collection of all :class:`~theng.data.pathfinder.Trigger` objects present in the Pathfinder simulation
        doors (DoorCollection): Collection of all :class:`~theng.data.pathfinder.Door` objects present in the Pathfinder simulation
        groups (GroupCollection): Collection of all :class:`~theng.data.pathfinder.Group` objects present in the Pathfinder simulation
        measurement_regions (MeasurementRegionCollection): Collection of all :class:`~theng.data.pathfinder.MeasurementRegion` objects present in the Pathfinder simulation
        occupants (OccupantCollection): Collection of all :class:`~theng.data.pathfinder.Occupant` objects present in the Pathfinder simulation
        rooms (RoomCollection): Collection of all :class:`~theng.data.pathfinder.Room` objects present in the Pathfinder simulation
        summary (Summary): Summary object that presents the data of the _summary.json file
        targets (TargetCollection): Collection of all :class:`~theng.data.pathfinder.Target` objects present in the Pathfinder simulation
        
        _triggers (Dict[str, Trigger]): Temp
        _doors (Dict[str, Door]): Temp
        _groups (Dict[str, Group]): Temp
        _measurement_regions (Dict[str, MeasurementRegion]): Temp
        _occupants (Dict[str, Occupant]): Temp
        _rooms (Dict[str, Room]): Temp
        _targets (Dict[str, Target]): Temp
    """
    
    def __init__(self, results_file_name: str):
        super().__init__(results_file_name, ".pth")
        self._triggers: Dict[str, Trigger] = {}
        self._doors: Dict[str, Door] = {}
        self._groups: Dict[str, Group] = {}
        self._measurement_regions: Dict[str, MeasurementRegion] = {}
        self._occupants: Dict[str, Occupant] = {}
        self._rooms: Dict[str, Room] = {}
        self._targets: Dict[str, Target] = {}
        
        self.triggers: TriggerCollection = TriggerCollection([])
        self.doors: DoorCollection = DoorCollection([])
        self.groups: GroupCollection = GroupCollection([])
        self.measurement_regions: MeasurementRegionCollection = MeasurementRegionCollection([])
        self.occupants: OccupantCollection = OccupantCollection([])
        self.rooms: RoomCollection = RoomCollection([])
        self.summary: Summary | None = None
        self.targets: TargetCollection = TargetCollection([])
        
        try:    
            self._detect_model_files()
            self._load()
        except Exception as e:
            self._log.exception(f'Error with PathfinderModel "{self.model_name}": {e}')
        
    def _detect_model_files(self) -> None:
        """Search for known model files and create DataFile objects for them
        
        Returns:
            None
        """
        for key, files in _known_output_files.items():
            for pair in files:
                file_type: str = pair[0]
                filename_format: str = pair[1]
                expected_filename_format: str = self.model_name + filename_format
                expected_file_location: str = os.path.join(os.getcwd(), expected_filename_format)
                if os.path.exists(expected_file_location):
                    try:
                        data_files_by_category: Optional[List] = self._results_data_files[key]
                    except KeyError as e:
                        self._results_data_files[key] = {}
                        
                    try:
                        data_files_by_type: Optional[List] = self._results_data_files[key][file_type]
                    except KeyError as e:
                        self._results_data_files[key][file_type] = []
                        data_files_by_type: Optional[List] = self._results_data_files[key][file_type]
                        
                    if file_type == "json":
                        data_files_by_type.append(JSONFile(expected_file_location))                        
                    else:  
                        data_files_by_type.append(DataFile(expected_file_location))
    
    def _load(self) -> None:
        """Loads results for this Model
        
        Returns:
            None
        """
        self._load_triggers()
        self._load_doors()
        self._load_groups()
        self._load_measurement_regions()
        self._load_rooms()
        self._load_targets()
        self._load_occupants() # Load last so occupants can reference other objects
        self._load_summary() # ^Last except for Summary. Occupants don't reference the Summary, but the Summary does reference Occupants
    
    def _load_triggers(self) -> None:
        """Load the raw _triggers.json data in to the PathfinderModel
        
        Returns:
            None
        """
        try:
            triggers_json: List[JSONFile] = self._results_data_files["triggers"]['json'][0]
            for key, data in triggers_json.data.items():
                self._triggers[key] = Trigger(name=key, data=data)
            self.triggers = TriggerCollection(self._triggers.values())
        except KeyError as e:
            if not get_arguments().quiet:
                self._log.warning(f'The "{self.get_model_name()}_triggers.json" file does not exist. Trigger data will not be available for this simulation. Make sure you have enabled JSON output in Pathfinder.')
    
    def _load_doors(self) -> None:
        """Load the _doors.json data in to the PathfinderModel
        
        Returns:
            None
        """
        try:
            doors_json: List[JSONFile] = self._results_data_files["doors"]['json'][0]
            print(doors_json)
            for key, data in doors_json.data.items():
                self._doors[key] = Door(name=key, data=data)
            self.doors = DoorCollection(self._doors.values())
        except KeyError as e:
            if not get_arguments().quiet:
                self._log.warning(f'The "{self.get_model_name()}_doors.json" file does not exist. Door data will not be available for this simulation. Make sure you have enabled JSON output in Pathfinder.')

    def _load_groups(self) -> None:
        """Load the _groups.json data in to the Pathfinder Model
        
        Returns:
            None
        """
        try:
            groups_json: Dict[str, Dict] = self._results_data_files["groups"]["json"][0]
            for group_id, data_dict in groups_json.data.items():
                self._groups[group_id] = Group(group_id=group_id, data=data_dict)
            self.groups = GroupCollection(self._groups.values())
        except KeyError as e:
            if not get_arguments().quiet:
                self._log.warning(f'The "{self.get_model_name()}_groups.json" file does not exist. Group data will not be available for this simulation. Make sure you have enabled JSON output in Pathfinder.')
     
    def _load_measurement_regions(self) -> None:
        """Load the _measurement-regions.json data in to the PathfinderModel
        
        Returns:
            None
        """
        try:
            regions_json: List[JSONFile] = self._results_data_files["measurement_regions"]['json'][0]
            for key, data in regions_json.data.items():
                self._measurement_regions[key] = MeasurementRegion(name=key, data=data)
            self.measurement_regions = MeasurementRegionCollection(self._measurement_regions.values())
        except KeyError as e:
            if not get_arguments().quiet:
                self._log.warning(f'The "{self.get_model_name()}_measurement-regions.json" file does not exist. Measurement Region data will not be available for this simulation. Make sure you have enabled JSON output in Pathfinder.')
       
    def _load_occupants(self) -> None:
        """Load raw _occupants.json, _occupant_params.json, and _occupants_detailed.json data in to the PathfinderModel
        
        Returns:
            None
        """
        occupants_json: Optional[List[JSONFile]] = None
        occupant_params_json: Optional[List[JSONFile]] = None
        occupants_detailed_json: Optional[List[JSONFile]] = None
        sd_accumulated_json: Optional[List[JSONFile]] = None
        sd_transient_json: Optional[List[JSONFile]] = None
        
        #Get general occupant data
        try:
            occupants_json: List[JSONFile] = self._results_data_files["occupants"]["json"][0]
        except KeyError as e:
            if not get_arguments().quiet:
                self._log.warning(f'The "{self.get_model_name()}_occupants.json" file does not exist. Occupant data will not be available for this simulation. Make sure you have enabled JSON output in Pathfinder.')
        
        #Get occupant param data
        try:
            occupant_params_json: List[JSONFile] = self._results_data_files["occupant_params"]["json"][0]
        except KeyError as e:
            pass
        
        #Get occupant detailed data
        try:
            occupants_detailed_json: List[JSONFile] = self._results_data_files["occupants_detailed"]["json"]
        except KeyError as e:
            if not get_arguments().quiet:
                self._log.warning(f'The detailed Occupant output JSON files do not exist. Detailed Occupant data will not be available for this simulation. Make sure you have enabled JSON output in Pathfinder.')
          
        #Extract Occupant detailed data from multi and single file outputs
        detailed_dict: Dict = {}
        if occupants_detailed_json != None:
            for detailed_file in occupants_detailed_json:
                for key, data in detailed_file.data.items():
                    detailed_dict[key] = data
                
        #Get Social Distancing Accumulated data
        try:
            sd_accumulated_json = self._results_data_files["social_distance_accumulated"]["json"][0]
        except KeyError as e:
            if not get_arguments().quiet:
                self._log.warning(f'The "{self.get_model_name()}_sd_accumulated.json" file does not exist. Accumulated Social Distancing data will not be available for this simulation. Make sure you have enabled JSON output in Pathfinder.')

        #Get Social Distancing Transient data
        try:
            sd_transient_json = self._results_data_files["social_distance_transient"]["json"][0]
        except KeyError as e:
            if not get_arguments().quiet:
                self._log.warning(f'The "{self.get_model_name()}_sd_transient.json" file does not exist. Transient Social Distancing data will not be available for this simulation. Make sure you have enabled JSON output in Pathfinder.')

          
        if occupants_json != None:
            for key, data in occupants_json.data.items():
                params: Optional[Dict] = occupant_params_json.data[key] if occupant_params_json != None else None
                details: Optional[Dict] = detailed_dict[key] if key in detailed_dict.keys() else None
                sd_accum: Optional[Dict] = sd_accumulated_json.data[key] if sd_accumulated_json != None else None
                sd_transient: Optional[Dict] = sd_transient_json.data[key] if sd_transient_json != None else None
                self._occupants[key] = Occupant(
                    occ_id=key, 
                    general_data=data, 
                    param_data=params, 
                    detailed_data=details, 
                    sd_accumulated_data=sd_accum, 
                    sd_transient_data=sd_transient, 
                    parent=self
                )
            self.occupants = OccupantCollection(self._occupants.values())
            for occ in self.occupants:
                occ.update_occupant_references()
        
    def _load_rooms(self) -> None:
        """Loads the raw _rooms.json data in to the PathfinderModel
        
        Returns:
            None
        """
        try:
            rooms_json: List[JSONFile] = self._results_data_files["rooms"]['json'][0]
            for key, data in rooms_json.data.items():
                self._rooms[key] = Room(name=key, data=data)
            self.rooms = RoomCollection(self._rooms.values())
        except KeyError as e:
            if not get_arguments().quiet:
                self._log.warning(f'The "{self.get_model_name()}_rooms.json" file does not exist. Room data will not be available for this simulation. Make sure you have enabled JSON output in Pathfinder.')
     
    def _load_summary(self) -> None:
        """Loads the raw _summary.json data in to the PathfinderModel
        
        Returns:
            None
        """
        try:
            summary_json: JSONFile = self._results_data_files["summary"]["json"][0]
            self.summary = Summary(parent=self, data=summary_json.data)
        except KeyError as e:
            if not get_arguments().quiet:
                self._log.warning(f'The "{self.get_model_name()}_summary.json" file does not exist. Summary data will not be available for this simulation. Make sure you have enabled JSON output in Pathfinder.')
      
    def _load_targets(self) -> None:
        """Load the raw _occtargets.json data in to the PathfinderModel
        
        Returns:
            None
        """
        try:
            targets_json: List[JSONFile] = self._results_data_files["targets"]['json'][0]
            for key, data in targets_json.data.items():
                self._targets[key] = Target(name=key, data=data)
            self.targets = TargetCollection(self._targets.values())
        except KeyError as e:
            if not get_arguments().quiet:
                self._log.warning(f'The "{self.get_model_name()}_targets.json" file does not exist. Targets data will not be available for this simulation. Make sure you have enabled JSON output in Pathfinder.')
    
    def take_screenshot(
        self, 
        ssview: str | List[str], 
        ssname: str | List[str] | None = None, sswidth: int = 800, 
        ssheight: int = 600, 
        sstime: float | int = 0
    ) -> List[str]:
        if 'nt' in sys.builtin_module_names:
            apps = [app for app in winapps.search_installed("pathfinder")]
            app = apps[0]
            
            results_path = f"{app.install_location}{os.sep}PathfinderResults.exe"
            return super()._take_screenshot(results_path, ssview, ssname, sswidth, ssheight, sstime)
        else:
            self._log.error("Tried to take a screenshot on a non-Windows platform")


from typing import Iterable, List, Union


from theng.data.collection import ADataCollection
from theng.data.pathfinder.behaviors.behavior import Behavior


class BehaviorCollection(ADataCollection):
    """Collection of :class:`~theng.data.pathfinder.Behavior` objects that provides array-like interface to make accessing Behavior data a little easier.
    
    Example:
        Elements in the Collection can be accessed in the following ways:
            - BehaviorCollection[str] => Gets the first Behavior object whose name matches the provided string.
            - BehaviorCollection[int] => Returns the Behavior object at the index specified by int.

    Arguments:
        behaviors (Union[Iterable, List]): Behavior objects to put in the collection
    """
    
    def __init__(self, *behaviors: Union[Iterable, List]):
        super().__init__(*behaviors)
        
    def __getitem__(self, key: Union[int, str]) -> Union[Behavior, List[Behavior]]:
        if type(key) == int:
            return self._elements[key]
        else:
            return next(behavior for behavior in self._elements if (behavior.name == key if type(behavior) == Behavior else behavior[0].name == key))
    


from typing import Dict
from theng.data.a_data_object import ADataObject

from theng.data.time_series import TimeSeries


class MeasurementRegion(ADataObject):
    """Represents a single Measurement Region in a Pathfinder simulation
    
    Args:
        name (str): The name of the Measurement Region.
        data (Dict): Raw data from the _measurement-regions.json file for this Region
            
    Attributes:
        name (str): The name of the Measurement Region
        density (TimeSeries[float]): TimeSeries Object that maps the Region's density measurements to Simulation time
        velocity (TimeSeries[float]): TimeSeries Object that maps the Region's velocity measurements to Simulation time
        seek_velocity (TimeSeries[float]): TimeSeries Object that maps the Region's seek_velocity measurements to Simulation time
            (If Seek Velocity output is enabled).
        count (TimeSeries[float]): TimeSeries Object that maps the Region's Occupant count measurements to Simulation time
    """
    
    def __init__(self, *, name: str, data: Dict) -> None:
        super().__init__()
        self.name = name
        self._data = data
        self.density: TimeSeries[float] = TimeSeries(time=[], values=[])
        self.velocity: TimeSeries[float] = TimeSeries(time=[], values=[])
        self.seek_velocity: TimeSeries[float] = TimeSeries(time=[], values=[])
        self.count: TimeSeries[int] = TimeSeries(time=[], values=[])
        
        self._load()
        
    def _load(self) -> None:
        self._load_safe_time_series(source_data=self._data, data_key='density', target='density')
        self._load_safe_time_series(source_data=self._data, data_key='velocity', target='velocity')
        self._load_safe_time_series(source_data=self._data, data_key='seekVelocity', target='seek_velocity')
        self._load_safe_time_series(source_data=self._data, data_key='count', target='count', data_type=int)




from abc import ABC
import os, logging, subprocess
from typing import Dict, List, Optional, Union

from theng.data.data_file import DataFile
from theng.data.json_file import JSONFile
from theng.plotting import return_image_plot


class Model(ABC):
    """Abstract class that represents a single simulation.

    Args:
        results_file_name (str): The basename of the results file
        extension (str): The standard extension for the simulator file. Used to construct the simulator file name
        
    Attributes:
        model_name (str): The basename of the simulator model file. None if the file is not detected.
        _log (logging.Logger): Convenience Logger
        _model_extension (str): See :py:attr:`~extension`
        _results_data_files (dict): Container dictionary for all detected results files
        _results_file_name (str): See :py:attr:`~results_file_name`
    """
    
    def __init__(self, results_file_name: str, extension: str):
        self._log: logging.Logger = logging.getLogger(f'theng.{self.__class__.__name__}')
        self._model_extension: str = extension
        self._results_data_files: Dict = {}
        self._results_file_name: str = results_file_name
        
        self.model_name: Optional[str] = None
        
        self.model_name = os.path.splitext(results_file_name)[0]
        self._basename = os.path.splitext(os.path.basename(results_file_name))[0]
    
    def get_model_name(self) -> Optional[str]:
        """Get the basename of the simulator file
        
        Returns:
            String basename if model file is valid, None if otherwise
        """
        return self.model_name
        
    def get_all_results_files(self) -> Dict[str, Dict[str, List[DataFile]]]:
        """Get all results data files for this model"""
        return self._results_data_files
    
    def get_results_json_files(self) -> Dict[str, JSONFile]:
        """Get all results data json files for this model"""
        return self._filter_results_files("json")
    
    def get_results_csv_files(self) -> Dict[str, DataFile]:
        """Get all results data csv files for this model"""
        return self._filter_results_files("csv")
    
    def get_results_txt_files(self) -> Dict[str, DataFile]:
        """Get all results data text files for this model"""
        return self._filter_results_files("txt")
                
    def _filter_results_files(self, extension_type: str) -> Dict[str, List[DataFile]]:
        """Get all detected results files for this model that match the given extension type
        
        Params:
            extension_type (str): The extension to filter by. Should match the type in self._results_data_files
        """
        files: Dict[str, List[DataFile]] = {}
        for descriptive_key, file_type_dict in self._results_data_files.items():
            for file_type, data_files in file_type_dict.items():
                if file_type == extension_type and data_files != None:
                    files[descriptive_key] = data_files
        return files
    
    def _take_screenshot(
        self, 
        results_path: str,
        ssview: Union[str, List[str]], 
        ssname: Union[str, List[str], None] = None, 
        sswidth: int = 800, 
        ssheight: int = 600, 
        sstime: Union[float, int] = 0
    ) -> List[str]:
        if type(ssname) == str:
            ssname = [ssname]
        if type(ssview) == str:
            ssview = [ssview]
        
        results: List[str] = []
        if not ssname:
            for view in ssview:
                name = f'{view}.png'
                command: List[str] = [results_path, '-screenshot', '-ssname', name, '-sswidth', str(sswidth), '-ssheight', str(ssheight), '-ssview', view, '-sstime', str(sstime), self._basename+".pfrv"]
                p = subprocess.Popen(command)
                p.wait()
                results.append(os.path.abspath(name))
                                
                return_image_plot(view, name)
        else:
            for name, view in zip(ssname, ssview):
                filename = f'{name}.png'
                command: List[str] = [results_path, '-screenshot', '-ssname', filename, '-sswidth', str(sswidth), '-ssheight', str(ssheight), '-ssview', view, '-sstime', str(sstime), self._basename+".pfrv"]
                p = subprocess.Popen(command)
                p.wait()
                results.append(os.path.abspath(filename))
                
                return_image_plot(name, filename)
        
        return results


from theng.data.pathfinder.occupants.occupant import Occupant, OccFrame, OccParams, SocialDistanceData, SocialDistanceFrame
from theng.data.pathfinder.occupants.occupant_collection import OccupantCollection


from typing import Iterable, List, Optional, Union
from theng.data.collection import ADataCollection
from theng.data.pathfinder.occupants.occupant import Occupant


class OccupantCollection(ADataCollection):
    """Collection of :class:`~theng.data.pathfinder.Occupant` objects that provides array-like interface to make accessing Occupant data a little easier.
    
    Example:
        Elements in the Collection can be accessed in the following ways:
            - OccupantCollection[str] => Gets the first Occupant object whose name matches the provided string.
            - OccupantCollection[int] => Returns the Occupant object at the index specified by int.

    Arguments:
        occs (Union[Iterable, List]): Occupant objects to put in the collection
    """
    
    def __init__(self, *occs: Union[Iterable, List]):
        super().__init__(*occs)
        
    def __getitem__(self,  key: Union[int, str]) -> Union[Occupant, List[Occupant]]:
        if type(key) == int:
            return self._elements[key]
        else:
            return next(occ for occ in self._elements if (occ.name == key))
        
    def get_by_id(self, id: int) -> Optional[Occupant]:
        """Search for and return the Occupant whose id attribute matches `id`

        Args:
            id (int): The id to search for
        """
        return next(occ for occ in self._elements if (occ.id == id))


from typing import Dict, List, Union
from theng.data.a_data_object import ADataObject

class Behavior(ADataObject):
    """Represents a single behavior in a Pathfinder simulation.

    Args:
        name (str): The full tree name of the behavior
        data (Dict): Raw data dict for the behavior from the _summary data file.
                    Dictionary should be in the following shape:
                    {
                        "time": {
                            "min": {
                                "occupant": Occupant,
                                "distance": float  
                            },
                            "max": {
                                "occupant": Occupant,
                                "distance": float       
                            },
                            "average": float,
                            "count": int,
                            "stdDev": float
                        }
                        "distance": {
                            "min": {
                                "occupant": Occupant,
                                "distance": float  
                            },
                            "max": {
                                "occupant": Occupant,
                                "distance": float       
                            },
                            "average": float,
                            "count": int,
                            "stdDev": float
                        }
                    }
        
    Attributes:
        name (str): The name of the behavior
        movement_distances (dict): Movement Distance data for this behavior
        completion_times (dict): Completion Time data for this behavior
        count (int): Number of Occupants that used this behavior
    """
    
    def __init__(self, *, name: str, data: Dict):
        super().__init__()
        self.name = name
        self._data = data
        
        self.movement_distances: dict = {}
        self.completion_times: dict = {}
        self.count: int = 0
        
        self._load()
        
    def _load(self) -> None:
        self.count = self._data["time"]["count"]
        self.completion_times = {
            "min": self._data["time"]["min"],
            "max": self._data["time"]["max"],
            "average": self._data["time"]["average"],
            "stdDev": self._data["time"]["stdDev"]
        }
        self.movement_distances = {
            "min": self._data["distance"]["min"],
            "max": self._data["distance"]["max"],
            "average": self._data["distance"]["average"],
            "stdDev": self._data["distance"]["stdDev"]
        }


from typing import Dict
from theng.data.a_data_object import ADataObject

from theng.data.time_series import TimeSeries


class Trigger(ADataObject):
    """Object representing an Trigger in the Simulation

    Args:
        name (str): The name specified in Pathfinder for this Trigger
        data (dict): Raw data from the _triggers.json file for this Trigger
        
    Attributes:
        name (str): See the `name` argument
        usage (TimeSeries[int]): TimeSeries object that maps the Triggers usage to Simulation Time
        _data (dict): Dictionary storing the raw data from the _triggers.json file
    """
    
    def __init__(self, *, name: str, data: Dict):
        super().__init__()
        self.name: str = name
        self._data: Dict = data
        
        self.usage: TimeSeries[int] = TimeSeries(time=[], values=[])
        
        self._load()
        
    def _load(self) -> None:
        self.usage = TimeSeries(time=[_ for _ in self._data.keys()], values=[int(_) for _ in self._data.values()])


from typing import Dict, List, Set, Union
from theng.data.a_data_object import ADataObject

from theng.data.time_series import TimeSeries


class Group(ADataObject):
    """Represents a single Group in a Pathfinder simulation
    
    Args:
        group_id (str): The group ID. Must be numeric. (The key in the _groups.json file)
        data (Dict): The data contained under the group ID in the _groups.json file
            
    Attributes:
        id (int): The group ID. Defaults to int(group_id)
        name (str): The name given to this group. Present if this group is predefined before the simulation.
        template (str): The name of the template use to generate this group. Present if this group is generated during the simulation.
        members (List[int]): List of IDs of all historical members of this group.
        member_history (TimeSeries[List[int]]): TimeSeries object storing membership information of this group over the Simulation
    """
    
    def __init__(self, *, group_id: str, data: Dict) -> None:
        super().__init__()
        self.id: int = int(group_id)
        self._data: Dict = data
        
        self.name: str = ""
        self.template: str = ""
        self.members: Union[List, Set] = []
        self.member_history: TimeSeries[List[int]] = TimeSeries(time=[], values=[])
        
        self._load()
        
    def _load(self) -> None:
        self._load_safe(source_data=self._data, data_key='name', target='name')
        self._load_safe(source_data=self._data, data_key='template', target='template')
        
        try:
            self._load_safe_time_series(source_data=self._data, data_key='members', target='member_history')
            for step in self.member_history.values:
                self.members += step
            self.members = set(self.members)
        except:
            pass


import os, sys
if 'nt' in sys.builtin_module_names:
    import winapps

from typing import List
from fdsreader import Simulation
from theng.models import Model

class FDSModel(Model, Simulation):
    """Represents a single FDS / PyroSim simulation and it's results.
    
    This class is a thin wrapper around the `Simulation <https://firedynamics.github.io/fdsreader/simulation.html>`_ class of fdsreader.

    Args:
        results_file_name (str): The basename of the .smv Results file generated by FDS
    """
    
    def __init__(self, results_file_name: str):
        
        Model.__init__(self, results_file_name, ".fds")
        Simulation.__init__(self, results_file_name)
        
    def take_screenshot(
        self, 
        app_name: str, 
        ssview: str | List[str], 
        ssname: str | List[str] | None = None, 
        sswidth: int = 800, 
        ssheight: int = 600, 
        sstime: float | int = 0
    ) -> List[str]:
        if 'nt' in sys.builtin_module_names:
            apps = [app for app in winapps.search_installed("pyrosim")]
            app = apps[0]
            
            results_path = f"{app.install_location}{os.sep}PyroSimResults.exe"
            return super()._take_screenshot(results_path, ssview, ssname, sswidth, ssheight, sstime)
        else:
            self._log.error("Tried to take a screenshot on a non-Windows platform")


"""Module to parse arguments passed by Thunderhead Results"""
import argparse
import sys, os
from typing import Optional

arguments: Optional[argparse.Namespace] = None
_argument_parser : argparse.ArgumentParser | None = None

if 'sphinx' not in sys.argv[0]:
    _argument_parser : argparse.ArgumentParser = argparse.ArgumentParser(add_help=False)

    _argument_parser.add_argument('--path', nargs="+", dest="pathfinder", help="Names of Pathfinder Results .pfr files")
    _argument_parser.add_argument('--fds', nargs="+", dest="pyrosim", help="Names of PyroSim Results .smv files")
    _argument_parser.add_argument("--ventus", nargs="+", dest="ventus", help="Names of Ventus Results files")
    _argument_parser.add_argument('-q', '--quiet', action="store_true", dest="quiet", help="Suppresses low level built-in warnings of the package")
    _argument_parser.add_argument('-h', '--help', action='help', default='==SUPPRESS==', help=('show this help message and exit'))

    arguments = _argument_parser.parse_known_args()[0]
    
def add_model_paths(*paths):
    """Adds additional paths to the list of arguments for later parsing by the package
    
    Arguments:
        paths (List[str]): Series or list of paths to add to the arguments
    
    Returns:
        The parsed arguments generated by the arguments parser
    """
    pfr = []
    smv = []
    vnt = []
    
    for path in paths:
        
        # This allows for a list of paths to be provided to the functions
        items = []
        if isinstance(path, list):
            items = [item for item in path]
        else:
            items = [path]
        for item in items:
            match os.path.splitext(item)[1]:
                case ".pfr":
                    pfr.append(item)
                case ".smv":
                    smv.append(item)
                case ".vnt":
                    vnt.append(item)
                
    if pfr != []:
        sys.argv.extend(["--path"])
        sys.argv.extend(x for x in pfr)
    if smv != []:
        sys.argv.extend(["--fds"])
        sys.argv.extend(x for x in smv)
    if vnt != []:
        sys.argv.extend(["--ventus"])
        sys.argv.extend(x for x in vnt)
        
    global arguments
    arguments = _argument_parser.parse_known_args()[0]
    return arguments

def suppress_warnings():
    sys.argv.extend(["-q"])
    global arguments
    arguments = _argument_parser.parse_known_args()[0]
    return arguments

def get_arguments():
    global arguments
    return arguments


"""The ``data`` package defines data management classes for use in extensions of :py:class:`theng.models.Model`.
It also contains subpackages for specific implementation of data management classes for supported simulators.

----

"""
from theng.data.data_file import DataFile
from theng.data.json_file import JSONFile
from theng.data.a_data_object import ADataObject
from theng.data.collection import ADataCollection
from theng.data.point import Point
from theng.data.time_series import TimeSeries
from theng.data.vector import Vector

__all__ = ["ADataObject", "ADataCollection", "Point", "Vector", "TimeSeries"]


from typing import Iterable, List, Union
from theng.data.collection import ADataCollection
from theng.data.pathfinder.tags.tag import Tag


class TagCollection(ADataCollection):
    """Collection of :class:`~theng.data.pathfinder.Tag` objects that provides array-like interface to make accessing Tag data a little easier.
    
    Example:
        Elements in the Collection can be accessed in the following ways:
            - TagCollection[str] => Gets the first Tag object whose name matches the provided string.
            - TagCollection[int] => Returns the Tag object at the index specified by int.

    Arguments:
        tags (Union[Iterable, List]): Tag objects to put in the collection
    """
    
    def __init__(self, *tags: Union[Iterable, List]):
        super().__init__(*tags)
        
    def __getitem__(self, key: Union[int, str]) -> Union[Tag, List[Tag]]:
        if type(key) == int:
            return self._elements[key]
        else:
            return next(tag for tag in self._elements if (tag.name == key if type(tag) == Tag else tag[0].name == key))


import os, json

from theng.data.data_file import DataFile

class JSONFile(DataFile):
    
    def __init__(self, filepath: str):
        """Class for reading JSON data files
        
        :invar filepath: The path to the JSON file
        :invar data: The data loaded from the JSON file
        """
        super().__init__(filepath)
        if os.path.splitext(filepath)[-1] != ".json":
            raise ValueError("JSONFile was created with a non-JSON filepath.")
        
        with open(filepath,  'r') as f:
            self.data = json.load(f) 


"""The pathfinder package provides members to interface with output files from the Thunderhead Engineering `Pathfinder <https://thunderheadeng.com/pathfinder>`_ simulator.
"""

from theng.data.pathfinder.triggers import Trigger, TriggerCollection
from theng.data.pathfinder.rooms import Room, RoomCollection
from theng.data.pathfinder.doors import Door, DoorCollection
from theng.data.pathfinder.groups import Group, GroupCollection
from theng.data.pathfinder.measurement_regions import MeasurementRegion, MeasurementRegionCollection
from theng.data.pathfinder.targets import Target, TargetCollection
from theng.data.pathfinder.occupants import Occupant, OccupantCollection, OccParams, OccFrame, SocialDistanceData, SocialDistanceFrame
from theng.data.pathfinder.summary import Summary
from theng.data.pathfinder.tags import Tag, TagCollection


from theng.models import Model

class VentusModel(Model):
    
    def __init__(self, results_file_name: str):
        """Class representing a single Ventus model file"""
        super().__init__(results_file_name, ".ventus")


from typing import Generic, List, TypeVar, Union, Self

T = TypeVar("T") 

class TimeSeries(Generic[T]):
    """Used to store Time Series based data about a Simulation
    
    Args:
        time (List[float]): Timesteps at which the data was written.
        values (List): A list of data corresponding to the Timesteps
            
    Attributes:
        time (List[float]): Timesteps at which the data was written.
        values (List[T]): A list of data corresponding to the Timesteps
    """
    
    def __init__(self, *, time: List[float], values: List[T]) -> None:
        self.index = 0
        self.time: List[float] = [float(x) for x in time]
        self.values: List[T] = values
        
    def __iter__(self):
        return TimeSeries(time=self.time, values=self.values)
    
    def __next__(self):
        next_obj = None
        while (next_obj is None):
            if self.index >= min(len(self.time), len(self.values)):
                raise StopIteration
            next_obj = [self.time[self.index], self.values[self.index]]
            self.index += 1
        return next_obj
    
    def __getitem__(self, key):
        if key <= len(self.time) and key <= len(self.values):
            return (self.time[key], self.values[key])
        raise IndexError
    
    """Filters the TimeSeries using a minimum and optional maximum time
    
    Args:
        min (float): The minimum time, inclusive, to filter by
        max (Optional[float, None]): The optional maximum time, inclusive, to filter by
        
    Returns:
        A new TimeSeries containing the filtered data.
    """
    def filter_by_time(self, min: float, max: Union[float, None] = None) -> Self:
        filtered_times = []
        filtered_values = []
        for [time, value] in self:
            if time >= min:
                if max == None or time <= max:
                    filtered_times.append(time)
                    filtered_values.append(value)
                    
        return TimeSeries(time=filtered_times, values=filtered_values)
    
    """Filters the TimeSeries using a minimum and optional maximum value
    
    Args:
        min (float): The minimum value, inclusive, to filter by
        max (Optional[float, None]): The optional maximum value, inclusive, to filter by
        
    Returns:
        A new TimeSeries containing the filtered data.
    """
    def filter_by_value(self, min: float, max: Union[float, None] = None) -> Self:
        filtered_times = []
        filtered_values = []
        for [time, value] in self:
            if value >= min:
                if  max == None or value <= max:
                    filtered_times.append(time)
                    filtered_values.append(value)
                    
        return TimeSeries(time=filtered_times, values=filtered_values)


from typing import Iterable, List, Union
from theng.data.collection import ADataCollection
from theng.data.pathfinder.measurement_regions.measurement_region import MeasurementRegion


class MeasurementRegionCollection(ADataCollection):
    """Collection of :class:`~theng.data.pathfinder.MeasurementRegion` objects that provides array-like interface to make accessing MeasurementRegion data a little easier.
    
    Example:
        Elements in the Collection can be accessed in the following ways:
            - MeasurementRegionCollection[str] => Gets the first MeasurementRegion object whose name matches the provided string.
            - MeasurementRegionCollection[int] => Returns the MeasurementRegion object at the index specified by int.

    Arguments:
        regions (Union[Iterable, List]): MeasurementRegion objects to put in the collection
    """
    
    def __init__(self, *regions: Union[Iterable, List]):
        super().__init__(*regions)
        
    def __getitem__(self,  key: Union[int, str]) -> Union[MeasurementRegion, List[MeasurementRegion]]:
        if type(key) == int:
            return self._elements[key]
        else:
            return next(region for region in self._elements if (region.name == key))


from theng.data.pathfinder.measurement_regions.measurement_region import MeasurementRegion
from theng.data.pathfinder.measurement_regions.measurement_region_collection import MeasurementRegionCollection


from typing import Iterable, List, Union


from theng.data.collection import ADataCollection
from theng.data.pathfinder.profiles.profile import Profile


class ProfileCollection(ADataCollection):
    """Collection of :class:`~theng.data.pathfinder.Profile` objects that provides array-like interface to make accessing Profile data a little easier.
    
    Example:
        Elements in the Collection can be accessed in the following ways:
            - ProfileCollection[str] => Gets the first Profile object whose name matches the provided string.
            - ProfileCollection[int] => Returns the Profile object at the index specified by int.

    Arguments:
        profiles (Union[Iterable, List]): Profile objects to put in the collection
    """
    
    def __init__(self, *profiles: Union[Iterable, List]):
        super().__init__(*profiles)
        
    def __getitem__(self, key: Union[int, str]) -> Union[Profile, List[Profile]]:
        if type(key) == int:
            return self._elements[key]
        else:
            return next(profile for profile in self._elements if (profile.name == key if type(profile) == Profile else profile[0].name == key))
    


from typing import Iterable, List, Union


from theng.data.collection import ADataCollection
from theng.data.pathfinder.rooms.room import Room


class RoomCollection(ADataCollection):
    """Collection of :class:`~theng.data.pathfinder.Room` objects that provides array-like interface to make accessing Room data a little easier.
    
    Example:
        Elements in the Collection can be accessed in the following ways:
            - RoomCollection[str] => Gets the first Room object whose name matches the provided string.
            - RoomCollection[int] => Returns the Room object at the index specified by int.

    Arguments:
        rooms (Union[Iterable, List]): Room objects to put in the collection
    """
    
    def __init__(self, *rooms: Union[Iterable, List]):
        super().__init__(*rooms)
        
    def __getitem__(self, key: Union[int, str]) -> Union[Room, List[Room]]:
        if type(key) == int:
            return self._elements[key]
        else:
            return next(room for room in self._elements if (room.name == key if type(room) == Room else room[0].name == key))
    


import os, logging
from typing import Dict, List, Optional

from theng.models import Model, PathfinderModel, FDSModel, VentusModel
from theng.args import get_arguments


_provided_models: dict = {
    "pathfinder": [],
    "pyrosim": [],
    "ventus": []
}

_known_file_extension_types: Dict[str, Model] = {
    ".pfr": PathfinderModel,
    ".smv": FDSModel,
    ".vnt": VentusModel
}

_LOGGER = logging.getLogger("theng")

def get_models() -> Dict[str, List[Model]]:
    """Get all models passed to the script by Results

    Returns:
        A Dictionary, keyed by the simulator name, containing all :py:class:`theng.models.Model` instances generated by data passed to the script.
        Keys are:
        
        - pathfinder
        - pyrosim
        - ventus (not currently supported)
    """ 
    get_pathfinder_models()
    get_pyrosim_models()
    # get_ventus_models()
    return _provided_models

def get_pathfinder_models() -> List[PathfinderModel]:
    """Gets all :py:class:`theng.models.PathfinderModel` instances generated by data passed to the script.

    Returns:
        A List of :py:class:`theng.models.PathfinderModel` instances generated by data passed to the script.
    """
    if (get_arguments().pathfinder != None):
        _provided_models["pathfinder"] = []
        for path in get_arguments().pathfinder:
            _provided_models["pathfinder"].append(_convert_to_model(path))
    return _provided_models["pathfinder"]

def get_pyrosim_models() -> List[FDSModel]:
    """Gets all :py:class:`theng.models.FDSModel` instances generated by data passed to the script.

    Returns:
        A List of :py:class:`theng.models.FDSModel` instances generated by data passed to the script.
    """
    if (get_arguments().pyrosim != None):
        _provided_models["pyrosim"] = []
        for path in get_arguments().pyrosim:
            _provided_models["pyrosim"].append(_convert_to_model(path))
    return _provided_models["pyrosim"]

def get_ventus_models() -> List[VentusModel]:
    """Ventus is not currently supported by theng.
    
    Gets all :py:class:`theng.models.VentusModel` instances generated by data passed to the script.

    Returns:
        A List of :py:class:`theng.models.VentusModel` instances generated by data passed to the script.
    """
    print("Ventus is not currently supported by theng")
    return []
    if (get_arguments().ventus != None):
        _provided_models["ventus"] = []
        for path in get_arguments().ventus:
            _provided_models["ventus"].append(_convert_to_model(path))
    return _provided_models["ventus"]
            
def _convert_to_model(results_file_name: str) -> Optional[Model]:
    """Factory to convert a results file (.pfr, .smv, .vnt) to an appropriate Model object
    
    Returns:
        A Model object instance if the results file has a valid associated model.
        None if otherwise, or if the extension is unknown.
    """
    ext: str = os.path.splitext(results_file_name)[-1]
    
    try:
        if ext in _known_file_extension_types.keys():
            return _known_file_extension_types[ext](results_file_name)
        return None
    except ValueError as e:
        _LOGGER.critical(f'{results_file_name} is not a supported results file type')
        return None     


"""Module to provide a number of pre-formatted plotting functions for use in Thunderhead Results"""

from enum import Enum
from typing import Iterable, List, Tuple, Union
import plotly.express as px
import plotly.graph_objects as go
from plotly.graph_objects import Figure

class PlotType(Enum):
    """Defines the valid types of plot file types"""
    HTML = 0
    PNG = 1
    JPG = 2
    JPEG = 3
    
def save_plot(fig: Figure, filename: str, plot_type: PlotType) -> None:
    """Saves the given figure as the provided PlotType and loads that file in to Thunderhead Results
    
    Args:
        fig (Figure): The Plotly Figure to save
        filename (str): The filename to save the Figure as
        type (FileType): The FileType to save the figure as
    """
    if plot_type == PlotType.HTML:
        path: str = filename + ".html"
        fig.write_html(path, include_mathjax='cdn')
        return_html_plot(fig.layout.title.text, path)
        
    elif plot_type == PlotType.PNG:
        path: str = filename + ".png"
        fig.write_image(path)
        return_image_plot(fig.layout.title.text, path)
        
    elif plot_type == PlotType.JPG:
        path: str = filename + ".jpg"
        fig.write_image(path)
        return_image_plot(fig.layout.title.text, path)
        
    elif plot_type == PlotType.JPEG:
        path: str = filename + ".jpeg"
        fig.write_image(path)
        return_image_plot(fig.layout.title.text, path)

def return_html_plot(name: str, filename: str) -> None:
    """Return an HTML plot object to Thunderhead Results for loading in the UI

    Args:
        name (str): The name of the plot to return and display
        filename (str): The path to the saved plot file
    """
    print(f'PLOTTITLE: {name}')
    print(f'HTMLPLOT: {filename}')
        
def return_image_plot(name: str, filename: str) -> None:
    """Return an image plot object to Thunderhead Results for loading in the UI

    Args:
        name (str): The name of the plot to return and display
        filename (str): The path to the saved plot file
    """
    print(f'PLOTTITLE: {name}')
    print(f'IMGPLOT: {filename}')

def line_plot(x_data: Union[List, Tuple[List]],
              y_data: Union[List, Tuple[List]], 
              x_title: str = None, 
              y_title: str = None, 
              plot_title: str = None, 
              width: int = 800, 
              height: int = 600, 
              legend_names: List[str] = None) -> Figure:
    """Utility plotting function for making line plots

    Args:
        x_data (Union[List, Tuple[List]]): 
            The data to plot on the X axis. Must be the same length and type as y_data.
            
            If a List, only one series will be plotted in the figure.
            If a tuple of lists, one series will be plotted for each pair of x_data and y_data.
        y_data (Union[List, Tuple[List]]): 
            The data to plot on the Y axis. Must be the same length and type as x_data.
            
            If a List, only one series will be plotted in the figure.
            If a tuple of lists, one series will be plotted for each pair of x_data and y_data.
        x_title (str, optional): Sets the X axis title. Defaults to None.
        y_title (str, optional): Sets the y axis title. Defaults to None.
        plot_title (str, optional): Sets the plot title. Defaults to None.
        width (int, optional): The width of the figure. Defaults to 800.
        height (int, optional): The height of the figure. Defaults to 600.
        legend_names (List[str], optional): Optional list of names to use when plotting multiple series of data. Defaults to a generator of "Series xxx" format.

    Raises:
        TypeError: If x_data and y_data are different types, not of type Tuple or List, or if they are a Tuple, but any of the elements in the tuple are not lists.
            Also raised if the legend_names List contains a Non-string.
        ValueError: If x_data, y_data, or legend_names (optional) are different lengths. 

    Returns:
        Figure: The final constructed figure
    """
    if type(x_data) != type(y_data):
        raise TypeError("X Data and Y Data are of incompatible type signatures for plotting.")
    elif type(x_data) not in (tuple, list):
        raise TypeError(f"Data provided is not an acceptable type. Expected Tuple or List, but got \"{type(x_data)}\"")
    elif len(x_data) != len(y_data):
        raise ValueError("X Data and Y Data are of different lengths.")
    
    if legend_names != None:
        if len(legend_names) != len(x_data):
            raise ValueError("Legend name generator does not supply a sufficient number of names")
    
    #Plot a single series of data
    if type(x_data) == list:
        data = {
            "x": x_data,
            "y": y_data
        }
        x_title = "X Data" if x_title == None else x_title
        y_title = "Y Data" if y_title == None else y_title
        plot_title = "Line Plot" if plot_title == None else plot_title
        
        fig: Figure = px.line(data, x="x", y="y")
        
    #Plot multiple series of data on the same figure
    elif type(x_data) == tuple:
        fig: Figure = Figure()
        if legend_names == None: 
            legend_names = [f'Series {i}' for i in range(0, len(x_data))]
        
        if legend_names != None:
            for x, y, name in zip(x_data, y_data, legend_names):
                if type(x) != list:
                    raise TypeError(f"Member of X Data tuple is not a valid type. Expected List, got \"{type(x)}\"")
                elif type(y) != list:
                    raise TypeError(f"Member of Y Data tuple is not a valid type. Expected List, got \"{type(y)}\"")
                elif type(name) != str:
                    raise TypeError(f"Name provided by legend_name is not a string: \"{name}\"")
                
                fig.add_trace(go.Scatter(
                    x=x,
                    y=y,
                    mode='lines',
                    name=name,
                    showlegend=True
                ))
    
    #Make it pretty by default
    fig.update_layout(
        title_text=plot_title,
        title_x=0.5,
        font_family="Courier New",
        font_color="#222222",
        paper_bgcolor="#cccccc",
        width=width,
        height=height,
        showlegend=True
    )
    fig.update_xaxes(title=x_title)
    fig.update_yaxes(title=y_title)
    return fig


from typing import Iterable, List, Union
from theng.data.collection import ADataCollection
from theng.data.pathfinder.doors import Door

class DoorCollection(ADataCollection):
    """Collection of :class:`~theng.data.pathfinder.Door` objects that provides array-like interface to make accessing Door data a little easier.
    
    Example:
        Elements in the Collection can be accessed in the following ways:
            - DoorCollection[str] => Gets the first Door object whose name matches the provided string.
            - DoorCollection[int] => Returns the Door object at the index specified by int.

    Arguments:
        doors (Union[Iterable, List]): Door objects to put in the collection
    """
    
    def __init__(self, *doors: Union[Iterable, List]):
        super().__init__(*doors)
        
    def __getitem__(self,  key: Union[int, str]) -> Union[Door, List[Door]]:
        if type(key) == int:
            return self._elements[key]
        else:
            return next(door for door in self._elements if (door.name == key if type(door) == Door  else  door[0].name == door))


class Vector():
    """Generic representation of a math Vector.
    
    Args:
        x (float): X component of the vector
        y (float): y component of the vector
        z (float): z component of the vector
        magnitude (float): Overall magnitude of the vector
        
    Attributes:
        x (float): X component of the vector
        y (float): Y component of the vector
        z (float): Z component of the vector
        magnitude (float): Overall magnitude of the vectory
    """
    #TODO Refactor magnitude to calculate the magnitude rather than displaying an assigned value.
    
    def __init__(self, x: float = float('inf'), y: float = float('inf'), z: float = float('inf'), magnitude: float = float('inf')):
        self.x: float = x
        self.y: float = y
        self.z: float = z
        self.magnitude: float = magnitude
        
    def __str__(self) -> str:
        return f'Vector(X={self.x}, Y={self.y}, Z={self.z}, magnitude={self.magnitude})'


from typing import Any, Dict, List, Optional, Union, TYPE_CHECKING

from theng.data.a_data_object import ADataObject
from theng.data.pathfinder import Trigger, Target
from theng.data.pathfinder.tags import Tag, TagCollection
from theng.data.point import Point
from theng.data.time_series import TimeSeries
from theng.data.vector import Vector
from theng.args import get_arguments


class Occupant(ADataObject):
    """Represents a single Occupant in a Pathfinder Simulation.

    Args:
        occ_id (str): Unique ID for this Occupant
        general_data (Dict): Dictionary of raw data from the _occupants.json file.
        param_data (Optional[Dict]): Optional Dictionary of raw data from the _occupant_params.json file
        detailed_data (Optional[Dict]): Optional Dictionary of raw data from the detailed occupant output files.
        sd_accumulated_data (Optional[Dict]): Optional Dictionary of raw data from the _sd_accumulated.json file.
        sd_transient_data (Optional[Dict]): Optional Dictionary of raw data from the _sd_transient.json file.
        parent (PathfinderModel): The PathfinderModel that this Occupant is a part of.
        
    Attributes:
        initial_params (OccParams): Initial parameters assigned to the Occupant in the Simulator. Contains the data of the _occupant_params files.
        history (Optional[TimeSeries[OccFrame]]): Optional TimeSeries detailed output data for this occupant. 
            None if the _occupant_detailed files are not present, or data is not present for this occupant.
        social_distancing (Optional[SocialDistanceData]): Optional SocialDistancingData for this occupant. 
            None if the _occupant_params files are not present, or data is not present for this occupant.
        
        id (int): A unique integer identifier for the occupant assigned by the simulator.
        name (str): The name given to the occupant in the user interface.
        exit_time (float): The time at which the Occupant went through and Exit Door.
            Default value is infinity. 
            This value represents that the Occupant has not exited the model.
        active_time (float): The amount of time the occupant was actively seeking a location in the model, such as a waypoint or exit.
        congestion_time_total (float):Total time the occupant spent moving at less than the congestion Velocity as specified in the simulation parameters (see `Section 14.1.2 of the Pathfinder User Manual <https://support.thunderheadeng.com/docs/pathfinder/latest/user-manual/#_output_parameters>`_).
        congestion_time_max_continuous (float): The maximum continuous amount of time the occupant spent at less than the congestion Velocity.
        congestion_time_level (float): Time spent by the occupant moving at less than congestion Velocity on level surfaces.
        congestion_time_stair (float): Time spent by the occupant moving at less than congestion Velocity on stairs.
        congestion_time_ramp (float): Time spent by the occupant moving at less than congestion Velocity on ramps.
        start_time (float): The time at which the occupant was generated. 
            For occupants pre-seeded in the model, this is always 0. 
            For occupants generated by an occupant source, this could be any time.
            The default value is infinity, representing that the Occupant did not report a start time.
        finish_time (float): The time at which the occupant left the simulation for any reason.
            The default value is infinity, representing that the occupant either did not exit the simulation, or did not report a finish time.
        distance (float): The total distance the occupant travelled during the simulation.
        tags (TagCollection): Collection of :class:`~theng.data.pathfinder.Tag` objects that were applied to this Occupant throughout the simulation.
        triggers_used (int): The total number of triggers used by the occupant during the simulation.
        triggers_usage_time (float): The amount of time spent by the occupant using triggers.
        targets_used (int): The total number of unique reserved Occupant Targets where the occupant waited during the simulation.
        target_usage_time (float): The total amount of time the occupant waited at a reserved Occupant Target.

        
        _general_data (Dict): Raw dict of data from the _occupants data files.
        _param_data (Dict): Raw dict of data from the _occupants_params data files.
        _detailed_data (Dict): Raw dict of data from the _occupants_detailed data files.
        _sd_accum_data (Dict): Raw dict of data from the _sd_accumulated data files.
        _sd_transient_data (Dict): Raw dict of data from the _sd_transient data files.
    """
    
    def __init__(
            self, 
            *, 
            occ_id: str, 
            general_data: Dict, 
            param_data: Optional[Dict], 
            detailed_data: Optional[Dict], 
            sd_accumulated_data: Optional[Dict], 
            sd_transient_data: Optional[Dict], 
            parent: Any = None
        ) -> None:
        super().__init__(parent=parent)
        self.id: int = int(occ_id)
        self._general_data: Dict = general_data
        self._param_data: Optional[Dict] = param_data
        self._detailed_data: Optional[Dict] = detailed_data
        self._sd_accum_data: Optional[Dict] = sd_accumulated_data
        self._sd_transient_data: Optional[Dict] = sd_transient_data
        
        # Loaded from General Occupant data
        self.name: str = ""
        self.exit_time: float = float('inf')
        self.active_time: float = 0.0
        self.congestion_time_total: float = 0.0
        self.congestion_time_max_continuous: float = 0.0
        self.congestion_time_level: float = 0.0
        self.congestion_time_stair: float = 0.0
        self.congestion_time_ramp: float = 0.0
        self.start_time: float = float('inf')
        self.finish_time: float = float('inf')
        self.distance: float = 0.0
        self.triggers_used: int = 0
        self.triggers_usage_time: float = 0.0
        self.targets_used: int = 0
        self.target_usage_time: float = 0.0
        self._tags_raw: Optional[Dict] = None
        self._tags: Dict[str, Tag] = {}
        self.tags: TagCollection = TagCollection([])
        
        # Loaded from Param data
        self.initial_params: OccParams = OccParams(param_data=None, parent=self)
        
        #Loaded from Detailed data
        self.history: Optional[TimeSeries[OccFrame]] = None
        
        #loaded from Transient and Accumulated Social Distancing Data
        self.social_distancing: Optional[SocialDistanceData] = None
        
        self._log.debug(f'Loading Occupant {self.id}')
        
        self._load_general_data()
        if self._param_data != None:
            self._load_param_data()
        if self._detailed_data != None:
            self._load_detailed_data()
        if self._sd_accum_data != None or self._sd_transient_data != None:
            self._load_sd_data()
        
    def _load_general_data(self) -> None:
        """Loads the data from the _occupants.json file"""
        self._log.debug(f'Loading General Occupant data for {self.id}')
        
        self._load_safe(source_data=self._general_data, data_key='name', target='name')
        self._load_safe(source_data=self._general_data, data_key='exitTime', target='exit_time')
        self._load_safe(source_data=self._general_data, data_key='activeTime', target='active_time')
        self._load_safe(source_data=self._general_data, data_key='congestionTimeTotal', target='congestion_time_total')
        self._load_safe(source_data=self._general_data, data_key='congestionTimeMaxContinuous', target='congestion_time_max_continuous')
        self._load_safe(source_data=self._general_data, data_key='congestionTimeLevel', target='congestion_time_level')
        self._load_safe(source_data=self._general_data, data_key='congestionTimeStair', target='congestion_time_stair')
        self._load_safe(source_data=self._general_data, data_key='congestionTimeRamp', target='congestion_time_ramp')
        self._load_safe(source_data=self._general_data, data_key='startTime', target='start_time')
        self._load_safe(source_data=self._general_data, data_key='finishTime', target='finish_time')
        self._load_safe(source_data=self._general_data, data_key='distance', target='distance')
        self._load_safe(source_data=self._general_data, data_key='numTriggersUsed', target='triggers_used')
        self._load_safe(source_data=self._general_data, data_key='triggersTime', target='triggers_usage_time')
        self._load_safe(source_data=self._general_data, data_key='numTargetsUsed', target='targets_used')
        self._load_safe(source_data=self._general_data, data_key='targetTime', target='target_usage_time')
        self._load_safe(source_data=self._general_data, data_key='tags', target='_tags_raw')
        
        if (self._tags_raw != None):
            self._load_tags_data()
            
    def _load_tags_data(self) -> None:
        """Format the loaded Tag data for this Occupant in to a TagCollection"""
        try:
            for key, data in self._tags_raw.items():
                self._tags[key] = Tag(name=key, data=data)
            self.tags = TagCollection(self._tags)
        except:
            self.tags = TagCollection([])
            if not get_arguments().quiet:
                self._log.warning(f'There was an error parsing tags for Occupant "{self.name}"')
        
    def _load_param_data(self) -> None:
        """Load the _occupant_params.json data"""
        self.initial_params = OccParams(param_data=self._param_data, parent=self)
        
    def _load_detailed_data(self) -> None:
        """load the detailed occupant output data"""        
        try:
            time: List[float] = []
            frames: List[OccFrame] = []
            for step, frame in self._detailed_data.items():
                time.append(float(step))
                frames.append(OccFrame(detailed_data=frame, parent=self))
            self.history = TimeSeries(time=time, values=frames)
        except:
            self.history = None
     
    def _load_sd_data(self) -> None:
        """Load social distancing data for this Occupant"""
        self.social_distancing = SocialDistanceData(sd_accum=self._sd_accum_data, sd_transient=self._sd_transient_data, parent=self)
        
    def update_occupant_references(self) -> None:
        """Update references to other Occupants in this Occupant object"""
        if self.social_distancing != None and type(self.social_distancing) == SocialDistanceData:
            self.social_distancing._update_occupant_references()
        
class OccParams(ADataObject):
    """Initial simulation parameters for an Occupant.

    Args:
        param_data (Optional[Dict]): Dict of raw param data for the Occupant.
        parent (Occupant): The Occupant associated with this data
        
    Attributes:
        parent (Occupant): The Occupant associated with this data
        profile (str): The name of the Profile assigned to this Occupant. 
        behavior (str): The name of the Behavior assigned to this Occupant.
        max_velocity (float): The maximum velocity allowed for this Occupant.
        shape (str): Name of the Occupant's assigned shape.
        room (str): The name of the room where the Occupant was created.
        position (Dict): Dict of X, Y, Z data of the Occupant's initial position.
        profile_props(Dict): Dictionary of profile properties. This is a temporary dictionary and will be changed in future versions
        
        _param_data (Optional[Dict]): Dict of raw param data for the Occupant.
    """
    
    def __init__(self, *, param_data: Optional[Dict], parent: Occupant):
        super().__init__()
        self._param_data: Optional[Dict] = param_data
        self.parent: Occupant = parent
        
        # Loaded from Occupant param data
        self.profile: str = ""
        self.behavior: str = ""
        self.max_velocity: float = float('inf')
        self.shape: str = {}
        self.room: str = ""
        self.position: Dict = {}
        self.profile_props: Optional[Dict] = None
        
        if self._param_data != None:
            self._load()
        
    def _load(self):
        """Loads the data from the _occupant_params.json file"""
        self._log.debug(f'Loading Occupant Parameter data for {self.parent.id}')
        
        self._load_safe(source_data=self._param_data, data_key="profile", target="profile")
        self._load_safe(source_data=self._param_data, data_key="behavior", target="behavior")
        self._load_safe(source_data=self._param_data, data_key="maxVelocity", target="max_velocity")
        self._load_safe(source_data=self._param_data, data_key="shape", target="shape")
        self._load_safe(source_data=self._param_data, data_key="room", target="room")
        self._load_safe(source_data=self._param_data, data_key="position", target="position")
        self._load_safe(source_data=self._param_data, data_key="profileProps", target="profile_props")
        
class OccFrame(ADataObject):
    """Represents the detailed data of an Occupant at a single time step.

    Args:
        detailed_data (Dict): The raw parsed detailed data for this frame of data.
        parent (Occupant): The :class:`~theng.data.pathfinder.Occupant` to which this frame of data applies.
        
    Attributes:
        parent (Occupant): The :class:`~theng.data.pathfinder.Occupant` to which this frame of data applies.
        active (bool): Whether the occupant is actively seeking an exit (True if they are seeking an exit and False if not).
        position (Point): The 3D location of the Occupant.
        velocity (Vector): The velocity of the Occupant.
        distance_traveled (float): The total distance the Occupant has travelled until this time in the simulation.
        location (str): The Occupant's current room.
        terrain (str): The terrain type of the Occupant's current room.
        trigger (Optioanl[Union[Trigger, str]]): The Trigger in use by the Occupant at the current time. None if no Trigger is in use.
        target (Optional[Union[Target, str]]): The Target in use by the Occupant at the current time. None if no Target is in use.
        co2_volume_fraction (Optional[float]): The current CO2 volume fraction exposure, if simulation is run with the FDS integration.
        co_volume_fraction (Optional[float]): The current CO volume fraction exposure, if simulation is run with the FDS integration.
        o2_volume_fraction (Optional[float]): The current O2 volume fraction exposure, if simulation is run with the FDS integration.
        temperature (Optional[float]): The current temperature exposure, if simulation is run with the FDS integration.
        soot_visibility (Optional[float]): The current visibility exposure, if simulation is run with the FDS integration.
        fed_total (Optional[float]): The current FED exposure, if simulation is run with the FDS integration.
    """
    
    def __init__(self, *, detailed_data: Dict, parent: Occupant) -> None:
        super().__init__()
        self._detailed_data: Dict = detailed_data
        self.parent: Occupant = parent
        
        self.active: bool = False
        self.position:  Point = Point()
        self.velocity: Vector = Vector()
        self.distance_traveled: float = float('inf')
        self.location: str = ""
        self.terrain: str = ""
        self.trigger: Optional[Union[Trigger, str]] = None #Trigger is loaded as a string from the file, then searched for in the model's TriggerCollection
        self.target: Optional[Union[Target, str]] = None #Target is loaded as a string from the file, then searched for in the model's TargetCollection
        self._tags_raw: List[str] = []
        self._tags: List[Tag] = []
        self.tags: TagCollection = TagCollection([])
        
        #FDS Integration data
        self.co2_volume_fraction: Optional[float] = None
        self.co_volume_fraction: Optional[float] = None
        self.o2_volume_fraction: Optional[float] = None
        self.temperature: Optional[float] = None
        self.soot_visibility: Optional[float] = None
        self.fed_total: Optional[float] = None
        
        self._load()
        
    def _load(self) -> None:
        self._log.debug(f'Loading Detailed Occupant data for {self.parent.id}')
        
        self._load_safe(source_data=self._detailed_data, data_key='isActive', target='active')
        self._load_safe(source_data=self._detailed_data, data_key='distance', target='distance_traveled')
        self._load_safe(source_data=self._detailed_data, data_key='location', target='location')
        self._load_safe(source_data=self._detailed_data, data_key='terrainType', target='terrain')
        self._load_safe(source_data=self._detailed_data, data_key='trigger', target='trigger')
        self._load_safe(source_data=self._detailed_data, data_key='tagsApplied', target='_tags')
        self._load_safe(source_data=self._detailed_data, data_key='carbonDioxideVolumeFraction', target='co2_volume_fraction')
        self._load_safe(source_data=self._detailed_data, data_key='carbonMonoxideVolumeFraction', target='co_volume_fraction')
        self._load_safe(source_data=self._detailed_data, data_key='oxygenVolumeFraction', target='o2_volume_fraction')
        self._load_safe(source_data=self._detailed_data, data_key='temperature', target='temperature')
        self._load_safe(source_data=self._detailed_data, data_key='sootVisibility', target='soot_visibility')
        self._load_safe(source_data=self._detailed_data, data_key='*fedTotal*', target='fed_total')
        
        
        try:
            position: Dict = self._detailed_data['position']
            self.position = Point(position['x'], position['y'], position['z'])
        except:
            self.position = Point()
        
        try:
            velocity: Dict = self._detailed_data['velocity']
            
            self.velocity = Vector(velocity['x'], velocity['y'], velocity['z'], velocity['magnitude'])
        except:
            self.velocity = Vector()
        
        try:
            self.trigger = self.parent.parent.triggers[self.trigger]
        except: 
            self.trigger = None
            
        try:
            for tag in self._tags_raw:
                self._tags.append(self.parent.tags[tag])
            self.tags = TagCollection(self._tags)
        except:
            self.tags = TagCollection([])
            
        try:
            self.target = self.parent.parent.targets[self.target]
        except: 
            self.target = None        

class SocialDistanceData(ADataObject):
    """Represents the Social Distance data for a single Occupant.

    Args:
        sd_accum (Dict): The raw parsed data for this Occupant from the sd_accumulated output file 
        sd_transient (Dict): The raw parted data for this Occupant from the sd_transient output file
        parent (Occupant): The :class:`~theng.data.pathfinder.Occupant` to which this data applies.
        
    Attributes:
        detection_radius (float): Distance used to calculate reference distance exposure.
        group_id (Optional[int]): Unique ID for occupant group, if present
        longest_occupant_id (Optional[int]): ID of occupant of greatest exposure (longest within distance SD)
        longest_occupant_name (Optional[str]): Name of occupant of greatest exposure (longest within distance SD)
        longest_occupant_time (Optional[float]): Time within distance SD of occupant of greatest exposure
        occupants_over_60 (List[Occupant]): List of Occupants within detection_radius for > 60 seconds.
        occupants_over_300 (List[Occupant]): List of Occupants within detection_radius for > 300 seconds.
        history (Optional[TimeSeries[SocialDistanceFrame]]): Optional TimeSeries object containing detailes transient SD data.
    """
    
    def __init__(self, *, sd_accum: Dict, sd_transient: Dict, parent: Occupant) -> None:
        super().__init__(parent)
        self._accum_data: Dict = sd_accum
        self._transient_data: Dict = sd_transient
        
        
        #Accumulated Data
        #TODO Verify this typing with a more detailed model
        self.detection_radius: float = float('inf')
        self.group_id: Optional[int] = None
        self.longest_occupant_id: Optional[int] = None
        self.longest_exposure: Optional[Dict] = None
        self._longest_exposure: Optional[Dict] = None;
        self.occupants_over_60: List[Occupant] = []
        self._occupants_over_60: Optional[Dict] = None
        self.occupants_over_300: List[Occupant] = []
        self._occupants_over_300: Optional[Dict] = None
        
        #Transient Data
        self.history: Optional[TimeSeries[SocialDistanceFrame]] = None
        
        self._load()
    
    def _load(self):
        if self._accum_data != None:
            self._load_safe(source_data=self._accum_data, data_key="detectionRadius", target="detection_radius")
            self._load_safe(source_data=self._accum_data, data_key="groupId", target="group_id")
            self._load_safe(source_data=self._accum_data, data_key="longestExposure", target="_longest_exposure")
            self._load_safe(source_data=self._accum_data, data_key="exposureOver60", target="_occupants_over_60")
            self._load_safe(source_data=self._accum_data, data_key="exposureOver300", target="_occupants_over_300")
        
        if self._transient_data != None:
            self.history = TimeSeries(
                time=[_ for _ in self._transient_data.keys()], 
                values=[SocialDistanceFrame(frame_data=_, parent=self.parent) for _ in self._transient_data.values()]
            )
            
    def _update_occupant_references(self):
        if self.history != None and type(self.history) == TimeSeries:
            for frame in self.history.values:
                frame._update_occupant_references()
                
        if self._occupants_over_60 != None:
            for occId in self._occupants_over_60["ids"]:
                occ: Occupant = self.parent.parent.occupants.get_by_id(int(occId))
                self.occupants_over_60.append(occ)
                
        if self._occupants_over_300 != None:
            for occId in self._occupants_over_300["ids"]:
                occ: Occupant = self.parent.parent.occupants.get_by_id(int(occId))
                self.occupants_over_300.append(occ)
                
        if self._longest_exposure != None:
            longestOcc = self.parent.parent.occupants.get_by_id(int(self._longest_exposure["id"]))
            time = self._longest_exposure["time"]
            self.longest_exposure = {
                "occupant": longestOcc,
                "time": time
            }
        
class SocialDistanceFrame(ADataObject):
    """Represents transient SD data at a single time step.

    Args:
        frame_data (Dict): An individual timestep frame of transient Social Distancing data.
        parent (Occupant): The :class:`~theng.data.pathfinder.Occupant` to which this frame of data applies.
        
    Attributes:
        group_id (Optional[int]): Unique ID for occupant group, if present
        closest_occupant (Optional[Dict]): Dictionary containing information about the closest Occupant to this Occupant.
                                           The "occupant" key will yield the Occupant object for the nearest Occupant.
                                           The "dist" key will yield the numerical distance as a float to the nearest Occupant.
        occupants_within_dist (Dict): Dictionary of Occupants within a given distance. Keys are '1.0', '2.0', and '3.0'.
    """
    
    def __init__(self, *, frame_data: Dict, parent: Occupant):
        super().__init__(parent=parent)
        self._data: Dict = frame_data
        
        self.group_id: Optional[int] = None
        self.closest_occupant: Optional[Dict] = None
        self._closest: Optional[Dict] = None
        self.occupants_within_dist: Dict = {}
        self._occs_within: Optional[Dict] = None
                
        self._load()
     
    def _load(self):
        self._load_safe(source_data=self._data, data_key="groupId", target="group_id")
        self._load_safe(source_data=self._data, data_key="closest", target="_closest")
        self._load_safe(source_data=self._data, data_key="occsWithin", target="_occs_within")
        
    def _update_occupant_references(self):
        if self._closest != None:
            self.closest_occupant = {
                "occupant": self.parent.parent.occupants.get_by_id(int(self._closest["id"])),
                "dist": float(self._closest["distance"])
            }
            
        if self._occs_within != None:
            for dist, obj in self._occs_within.items():
                distance: float = float(dist)
                self.occupants_within_dist[distance] = []
                occIds: List[float] = obj['ids']
                for occId in occIds:
                    self.occupants_within_dist[distance].append(self.parent.parent.occupants.get_by_id(int(occId)))


"""
.. seealso::
    Support for FDS is provided through a thin wrapper around the `fdsreader <https://github.com/FireDynamics/fdsreader>`_ package developed by `Forschungszentrum Jülich <https://www.fz-juelich.de>`_.
    The full API documentation is available `here <https://firedynamics.github.io/fdsreader/>`_.
    
    Special thanks to the developers for their hard work they've put in to that package.
    
"""

from theng.models.fds.fds import FDSModel


"""`theng </>`_ is a data processing toolset for common Life Safety simulators and a collection of utilities for displaying processed data in Thunderhead Results.
This toolset is designed for use in the Scripting Engine of the `Thunderhead Results Viewer <https://thunderheadeng.com>`_, however it can also be used in standalone Python scripts.

Attributes:
    log(Optional[logging.Logger]): A preconfigured logger that writes output to the script log file.
    VERSION(str): The version of the installed theng package
"""
import sys

from theng.main import *
from theng.logging import get_theng_logger, _initialize_logger
from theng.args import add_model_paths, suppress_warnings

log: Optional[logging.Logger] = None
if 'sphinx' not in sys.argv[0]:
    _initialize_logger()
    log = get_theng_logger(sys.argv[0])   
    
__all__ = ["get_models", "get_pathfinder_models", "get_pyrosim_models", "get_ventus_models", "get_theng_logger", "add_model_paths", "suppress_warnings"]

VERSION: str = "0.1.6"


from abc import ABC
from logging import Logger
from typing import Any, Dict
from theng.data.time_series import TimeSeries

from theng.logging import get_theng_logger
from theng.args import get_arguments

class ADataObject(ABC):
    
    def __init__(self, parent: Any = None):
        self.parent = parent
        self._log: Logger = get_theng_logger(str(self.__class__.__name__))
    
    def _load_safe(self, *, source_data: Dict, data_key: str, target: str, e_message: str = None, data_type: Any = None) -> None:
        """Safely load a value from a dictionary in to this class as an attribute with name defined by `target`

        Args:
            source_data (Dict): The data dictionary to load from
            data_key (str): The key in the dictionary to pull the value from
            target (str): The attribute name to set on this class
            e_message (str, optional): A message to warn in the log if the data_key cannot be accessed. Defaults to None.
            data_type (Any): The type to cast any dictionary values to. Defaults to None, which will just load the data as whatever type it already is.
        """
        try:
            val: Any = source_data[data_key]
            if val == "None" or val == "":
                val = None
            if data_type != None and val != None:
                self.__setattr__(target, data_type(val))
            else:
                self.__setattr__(target, val)
        except KeyError:
            if e_message != None and not get_arguments().quiet:
                self._log.warning(e_message)
                
    def _load_safe_time_series(self, *, source_data: Dict, data_key: str, target: str, e_message: str = None, data_type: Any = None) -> None:
        """Safely load a TimeSeries in to this class from a dictionary as an attribute with the name defined by `target`

        Args:
            source_data (Dict): The data dictionary to load from
            data_key (str): The key in the dictionary to pull the values from
            target (str): The attribute name to set on this class
            e_message (str, optional): A message to warn in the log if the data_key cannot be accessed. Defaults to None.
            data_type (Any): The type to cast any dictionary values to. Defaults to None, which will just load the data as whatever type it already is.
        """
        try:
            if data_type != None:
                self.__setattr__(target, TimeSeries(time=[_ for _ in source_data[data_key].keys()], values=[data_type(_) for _ in source_data[data_key].values()]))
            else:
                self.__setattr__(target, TimeSeries(time=[_ for _ in source_data[data_key].keys()], values=[_ for _ in source_data[data_key].values()]))
        except KeyError:
            if e_message != None and not get_arguments().quiet:
                self._log.warning(e_message)


from typing import Dict, Union
from theng.data.a_data_object import ADataObject

from theng.data.time_series import TimeSeries

#Guards against circular imports
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from theng.data.pathfinder.occupants.occupant import Occupant


class Room(ADataObject):
    """Represents a single room in a Pathfinder simulation.

    Args:
        name (str): The full tree name of the room
        data (Dict): Raw data dict for the room from the _rooms data file.
        
    Attributes:
        name (str): The full tree name of the room
        shortname (str): The last portion of the full tree name of the room.
        usage (TimeSeries[int]): Usage data for this door over time
        total_usage (int): Total number of Occupants that have used this room during the simulation. 
                            This data is loaded from the simulation's _summary.json file and is not applicable to every Room
        first_in_time (float): The time when this Room is first use by an Occupant in the simulation. 
                            This data is loaded from the simulation's _summary.json file and is not applicable to every Room
        last_out_time (float): The time when this Room is last used by an Occupant in the simulation. 
                            This data is loaded from the simulation's _summary.json file and is not applicable to every Room
        last_out_occupant (Union[Occupant, None]): The last Occupant to have used this room in the simulation.
                            This data is loaded from the simulation's _summary.json file and is not applicable to every Room
    """
    
    def __init__(self, *, name: str, data: Dict):
        super().__init__()
        self.name = name
        self.shortname = self.name.split('->')[-1]
        self._data = data
        
        self.usage: TimeSeries[int] = TimeSeries(time=[], values=[])
        self.total_usage: int = float('inf')
        self.first_in_time: float = float("inf")
        self.last_out_time: float = float("inf")
        self.last_out_occupant: Union["Occupant", None] = None
        
        self._load()
        
    def _load(self) -> None:
        self.usage = TimeSeries(time=[_ for _ in self._data.keys()], values=[int(_) for _ in self._data.values()])


"""The ``models`` package provides a generic :py:class:`theng.models.Model` class to represent simulation files, and also contains extending classes that are specific to each supported simulator.

----

"""
from theng.models.model import Model
from theng.models.pathfinder import PathfinderModel
from theng.models.ventus import VentusModel
from theng.models.fds import FDSModel

__all__ = ["Model", "PathfinderModel", "FDSModel", "VentusModel"]


from typing import Dict, List
from theng.data.a_data_object import ADataObject
from theng.data.pathfinder.profiles import ProfileCollection, Profile
from theng.data.pathfinder.behaviors import BehaviorCollection, Behavior

class Summary(ADataObject):
    """Placeholder for Simulation Summary data.

    Args:
        data (Data): data
        
    Attributes:
        cpu_time (float):
        startup_time (float):
        components_all (int):
        simulation (str):
        version (str):
        mode (str):
        triangles (int):
        total_occupants (int):
        components_door (int):
        profiles (ProfileCollection):
        behaviors (BehaviorCollection):
        completion_times (dict):
        movement_distances (dict):
        _completion_times_profile (dict):
        _completion_times_behavior (dict):
        _completion_times_all (dict):
        _movement_distances_profile (dict):
        _movement_distances_behavior (dict):
        _movement_distances_all (dict):
        _room_usage (List):
        _door_flow_rates (List):
    """
    
    def __init__(self, parent, data: Dict) -> None:
        super().__init__(parent=parent)
        self._data = data
        
        self.cpu_time: float = float('inf')
        self.startup_time: float = float('inf')
        self.components_all: int = 0
        self.simulation: str = ""
        self.version: str = ""
        self.mode: str = ""
        self.triangles: int = 0
        self.total_occupants: int = 0
        self.components_doors: int = 0
        self.profiles = ProfileCollection([])
        self.behaviors = BehaviorCollection([])
        self.completion_times: dict = {}
        self.movement_distances: dict = {}
        
        self._completion_times_profile: List = []
        self._completion_times_behavior: List = []
        self._completion_times_all: dict = {}
        self._movement_distance_profile: List = []
        self._movement_distances_behavior: List = []
        self._movement_distances_all: dict = {}
        
        self._room_usage: List = []
        self._door_flow_rates: List = []
        
        self._load()
    
    def _load(self) -> None:
        self._load_safe(source_data=self._data, data_key="cpu_time", target="cpu_time")
        self._load_safe(source_data=self._data, data_key="startup_time", target="startup_time")
        self._load_safe(source_data=self._data, data_key="components_all", target="components_all", data_type=int)
        self._load_safe(source_data=self._data, data_key="simulation", target="simulation")
        self._load_safe(source_data=self._data, data_key="version", target="version")
        self._load_safe(source_data=self._data, data_key="mode", target="mode")
        self._load_safe(source_data=self._data, data_key="triangles", target="triangles")
        self._load_safe(source_data=self._data, data_key="total_occupants", target="total_occupants")
        self._load_safe(source_data=self._data, data_key="components_doors", target="components_doors")
        self._load_safe(source_data=self._data, data_key="completion_times_all", target="_completion_times_all")
        self._load_safe(source_data=self._data, data_key="completion_times_profile", target="_completion_times_profile")
        self._load_safe(source_data=self._data, data_key="completion_times_behavior", target="_completion_times_behavior")
        self._load_safe(source_data=self._data, data_key="movement_distances_all", target="_movement_distances_all")
        self._load_safe(source_data=self._data, data_key="movement_distance_profile", target="_movement_distance_profile")
        self._load_safe(source_data=self._data, data_key="movement_distances_behavior", target="_movement_distances_behavior")
        self._load_safe(source_data=self._data, data_key="room_usage", target="_room_usage")
        self._load_safe(source_data=self._data, data_key="door_flow_rates", target="_door_flow_rates")
        
        self._load_completion_times()
        self._load_movement_distances()
        self._load_profiles()
        self._load_behaviors()
        self._load_room_usage()
        self._load_door_flow_rates()
        
        
    def _load_profiles(self) -> None:
        profile_data: dict[str, dict] = {}
        for entry in self._movement_distance_profile:
            profile_data[entry["profile"]] = {}
            profile_data[entry["profile"]]["distance"] = {
                "min": {
                    "occupant": self.parent.occupants[entry["min"]["name"]],
                    "distance": float(entry["min"]["distance"])   
                },
                "max": {
                    "occupant": self.parent.occupants[entry["max"]["name"]],
                    "distance": float(entry["max"]["distance"])       
                },
                "average": float(entry["avg"]),
                "count": int(entry["count"]),
                "stdDev": float(entry["stdDev"])
        }
            
        for entry in self._completion_times_profile:
            profile_data[entry["profile"]]["time"] = {
                "min": {
                    "occupant": self.parent.occupants[entry["min"]["name"]],
                    "time": float(entry["min"]["time"])   
                },
                "max": {
                    "occupant": self.parent.occupants[entry["max"]["name"]],
                    "time": float(entry["max"]["time"])       
                },
                "average": float(entry["avg"]),
                "count": int(entry["count"]),
                "stdDev": float(entry["stdDev"])
            }
            
        profiles: List[Profile] = [Profile(name=name, data=data) for name, data in profile_data.items()]
        self.profiles = ProfileCollection(profiles)
                
    def _load_behaviors(self) -> None:
        behavior_data: dict[str, dict] = {}
        for entry in self._movement_distances_behavior:
            behavior_data[entry["behavior"]] = {}
            behavior_data[entry["behavior"]]["distance"] = {
                "min": {
                    "occupant": self.parent.occupants[entry["min"]["name"]],
                    "distance": float(entry["min"]["distance"])   
                },
                "max": {
                    "occupant": self.parent.occupants[entry["max"]["name"]],
                    "distance": float(entry["max"]["distance"])       
                },
                "average": float(entry["avg"]),
                "count": int(entry["count"]),
                "stdDev": float(entry["stdDev"])
        }
            
        for entry in self._completion_times_behavior:
            behavior_data[entry["behavior"]]["time"] = {
                "min": {
                    "occupant": self.parent.occupants[entry["min"]["name"]],
                    "time": float(entry["min"]["time"])   
                },
                "max": {
                    "occupant": self.parent.occupants[entry["max"]["name"]],
                    "time": float(entry["max"]["time"])       
                },
                "average": float(entry["avg"]),
                "count": int(entry["count"]),
                "stdDev": float(entry["stdDev"])
            }
            
        behaviors: List[Behavior] = [Behavior(name=name, data=data) for name, data in behavior_data.items()]
        self.behaviors = BehaviorCollection(behaviors)
                
    def _load_completion_times(self) -> None:
        if self._data:
            self.completion_times = {
                "min": {
                    "occupant": self.parent.occupants[self._completion_times_all["min"]["name"]],
                    "time": float(self._completion_times_all["min"]["time"])   
                },
                "max": {
                    "occupant": self.parent.occupants[self._completion_times_all["max"]["name"]],
                    "time": float(self._completion_times_all["max"]["time"])       
                },
                "average": float(self._completion_times_all["average"]),
                "stdDev": float(self._completion_times_all["stdDev"])
            }
        
    def _load_movement_distances(self) -> None:
        if self._data:
            self.movement_distances = {
                "min": {
                    "occupant": self.parent.occupants[self._movement_distances_all["min"]["name"]],
                    "distance": float(self._movement_distances_all["min"]["distance"])   
                },
                "max": {
                    "occupant": self.parent.occupants[self._movement_distances_all["max"]["name"]],
                    "distance": float(self._movement_distances_all["max"]["distance"])       
                },
                "average": float(self._movement_distances_all["average"]),
                "stdDev": float(self._movement_distances_all["stdDev"])
            }
        
        
    """
    Loads data from the Summary file in to the Room objects for the relevant Rooms
    """
    def _load_room_usage(self) -> None:
        for room_entry in self._room_usage:
            name = room_entry["room"]
            total_usage = room_entry["total_use"]
            first_in = float(room_entry["first_in"])
            last_out = float(room_entry["last_out"])
            last_out_name: str = room_entry["last_out_name"].strip()
            
            room = self.parent.rooms[name]
            last_out_occ = None
            if last_out_name != "":
                last_out_occ = self.parent.occupants[last_out_name]
            
            room.total_usage = total_usage
            room.first_in_time = first_in
            room.last_out_time = last_out
            room.last_out_occupant = last_out_occ
                
    """
    Loads data from the Summary file in to the Door objects for the relevant Doors
    """
    def _load_door_flow_rates(self) -> None:
        for door_entry in self._door_flow_rates:
            total_usage = door_entry["total_use"]
            name = door_entry["door"]
            first_in = float(door_entry["first_in"])
            last_out = float(door_entry["last_out"])
            
            flow_avg_raw = door_entry["flow_avg"]
            if flow_avg_raw != "":
                flow_avg = float(flow_avg_raw)
            else:
                flow_avg = 0.0
                
            last_out_name = door_entry["last_out_name"].strip()
            last_out_occ = None
            if last_out_name != "":
                last_out_occ = self.parent.occupants[last_out_name]
                
            door = self.parent.doors[name]
            door.total_usage = total_usage
            door.first_in_time = first_in
            door.last_out_time = last_out
            door.flow_avg = flow_avg
            door.last_out_occupant = last_out_occ                


from typing import Dict, List, Union
from theng.data.a_data_object import ADataObject

from theng.data.time_series import TimeSeries


class Target(ADataObject):
    """Represents a single Occupant Target in a Pathfinder simulation
    
    Args:
        name (str): The name of the Target.
        data (Dict): Raw data from the _occtargets output file for this Target
            
    Attributes:
        name (str): The name of the Target
        is_in_use (TimeSeries[bool]): TimeSeries representing the usage of this Target over time.
        reservation_ids_history (TimeSeries[int]): TimeSeries representing the IDs of Occupants using this Target over time.
        reservation_name_history (TimeSeries[str]): TimeSeries representing the names of Occupants using this Target over time.
    """
    
    def __init__(self, *, name: str, data: Dict):
        super().__init__()
        self.name: str = name
        self._data: Dict = data
        
        self.is_in_use: TimeSeries[bool] = TimeSeries(time=[], values=[])
        self.reservation_ids_history: TimeSeries[int] = TimeSeries(time=[], values=[])
        self.reservation_names_history: TimeSeries[str] = TimeSeries(time=[], values=[])
        
        self._load()
        
    def _load(self) -> None:
        self._load_safe_time_series(source_data=self._data, data_key='using', target='is_in_use')
        self._load_safe_time_series(source_data=self._data, data_key='reservedById', target='reservation_ids')
        self._load_safe_time_series(source_data=self._data, data_key='reservedBy', target='reservation_names')


from typing import Iterable, List, Union
from theng.data.collection import ADataCollection
from theng.data.pathfinder.triggers import Trigger

class TriggerCollection(ADataCollection):
    """Collection of :class:`~theng.data.pathfinder.Trigger` objects that provides array-like interface to make accessing trigger data a little easier.
    
    Example:
        Elements in the Collection can be accessed in the following ways:
            - TriggerCollection[str] => Gets the first Trigger object whose name matches the provided string.
            - TriggerCollection[int] => Returns the Trigger object at the index specified by int.

    Arguments:
        triggers (Union[Iterable, List]): Trigger objects to put in the collection
    """
    
    def __init__(self, *triggers: Union[Iterable, List]):
        super().__init__(*triggers)
        
    def __getitem__(self,  key: Union[int, str]) -> Union[Trigger, List[Trigger]]:
        if type(key) == int:
            return self._elements[key]
        else:
            return next(attr for attr in self._elements if (attr.name == key if type(attr) == Trigger  else  attr[0].name == attr))


import logging, sys, os, datetime

def _initialize_logger() -> None:
    """Initializes the theng logger"""
    _LOGGER = logging.getLogger("theng")
    _LOGGER.setLevel(logging.INFO)
    path = f'{os.getcwd()}/{(os.path.splitext(os.path.basename(sys.argv[0]))[0])}_script_logging.txt'
    
    if (os.path.exists(path)): os.remove(path)
    
    _file_handler = logging.FileHandler(path)
    _file_formatter = logging.Formatter(fmt="%(name)s :: %(levelname)s :: %(message)s")
    _file_handler.setFormatter(_file_formatter)
    
    _LOGGER.addHandler(_file_handler)
    _LOGGER.info(f"Logging initialized at {datetime.datetime.now().strftime('%H:%M:%S on %m/%d/%Y')}")
    
def get_theng_logger(file: str) -> logging.Logger:
    """Gets the theng logger for the given python file
    
    :param file: The __file__ variable for the python file
    
    Returns:
        A Logger instance using a name with format `theng.file`
    """
    return logging.getLogger(f'theng.{os.path.basename(os.path.splitext(file)[0])}')


from enum import Enum


class Direction(Enum):
    BIDIRECTIONAL = None
    Y_POSITIVE = "[+Y]"
    Y_NEGATIVE = "[-Y]"
    X_POSITIVE = "[+X]"
    X_NEGATIVE = "[-X]"
    Z_POSITIVE = "[+Z]"
    Z_NEGATIVE = "[-Z]"
    
    def __str__(self) -> str:
        if self.value == None:
            return "BiDirectional"
        return self.value.replace("[", "").replace("]", "")
    
    def __eq__(self, other):
        if other == None:
            if self.value == None:
                return True
        
        if isinstance(other, str): 
            return self.value == other
        
        if isinstance(other, Direction):
            return self is other
        
        return False


from typing import Iterable, List, Union
from theng.data.collection import ADataCollection
from theng.data.pathfinder.targets import Target

class TargetCollection(ADataCollection):
    """Collection of :class:`~theng.data.pathfinder.Target` objects that provides array-like interface to make accessing Target data a little easier.
    
    Example:
        Elements in the Collection can be accessed in the following ways:
            - TargetCollection[str] => Gets the first Target object whose name matches the provided string.
            - TargetCollection[int] => Returns the Target object at the index specified by int.

    Arguments:
        targets (Union[Iterable, List]): Target objects to put in the collection
    """
    
    def __init__(self, *targets: Union[Iterable, List]):
        super().__init__(*targets)
        
    def __getitem__(self,  key: Union[int, str]) -> Union[Target, List[Target]]:
        if type(key) == int:
            return self._elements[key]
        else:
            return next(target for target in self._elements if (target.name == key if type(target) == Target  else  target[0].name == target))


import re, logging
from typing import List, Optional, Union, Dict
from theng.data.a_data_object import ADataObject

from theng.data.pathfinder.direction import Direction
from theng.data.time_series import TimeSeries

#Guards against circular imports
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from theng.data.pathfinder.occupants.occupant import Occupant

class Door(ADataObject):
    """Data class that represents a single door in a Pathfinder Simulation
    
    Args:
        name (str): The complete full name of the Door in the Pathfinder model in the form of "Floor -> group -> Doorname"
        data (Dict[str, Union[List[Union[float, int]], Dict[str, List[Union[float, int]]]]]): The raw data dictionary for the Door
            
    Attributes:
        name (str): The complete full name of the Door in the Pathfinder model. e.g. "Floor -> group -> Doorname"
        shortname (str): The last portion of the name attribute. e.g. "Doorname"
        usage (TimeSeries[int]): TimeSeries usage data for the door throughout the simulation
        width (TimeSeries[float]): TimeSeries width data for the door throughout the simulation
        boundary (TimeSeries[float]): TimeSeries boundary width data for the door throughout the simulation
        queueing_usage (TimeSeries[int]): TimeSeries queueing usage data for the door throughout the simulation
        direction (Direction): Direction.BIDIRECTIONAL if this door is not oneway,  one of the other Direction values if otherwise
        total_usage (int): Total number of Occupants that used this door during the simulation.
                            This value is loaded from the _summary.json file and may not be relevant to every Door entry.
        first_in_time (float): Time when an Occupant first used this door during the simulation.
                            This value is loaded from the _summary.json file and may not be relevant to every Door entry.
        last_out_time (float): Time when an Occupant last used this door during the simulation.
                            This value is loaded from the _summary.json file and may not be relevant to every Door entry.
        flow_avg (float): Average flow rate for this door over the entire simulation
                            This value is loaded from the _summary.json file and may not be relevant to every Door entry.
        last_out_occupant (Union[Occupant, None]): The last Occupant to use this door during the simulation
                            This value is loaded from the _summary.json file and may not be relevant to every Door entry.
    """
    
    def __init__(self, *, name: str, data: Dict[str, Union[List[Union[float, int]], Dict[str, List[Union[float, int]]]]]):
        super().__init__()
        self.name: str = name
        self.shortname: str = name.split("->")[-1]
        self._data: Dict[str, Union[List[Union[float, int]], Dict[str, List[Union[float, int]]]]] = data
        
        self.usage: TimeSeries[int] = TimeSeries(time=[], values=[])
        self.width: TimeSeries[float] = TimeSeries(time=[], values=[])
        self.boundary: TimeSeries[float] = TimeSeries(time=[], values=[])
        self.queueing_usage: TimeSeries[int] = TimeSeries(time=[], values=[])
        self.direction: Direction = Direction.BIDIRECTIONAL
        
        if self.name == "exited" or self.name == "remaining":
            self._load_summary_data()
        else:
            self._load_individual_door_data()
            
        self.total_usage: int = 0
        self.first_in_time: float = float('inf')
        self.last_out_time: float = float('inf')
        self.flow_avg: float = float('inf')
        self.last_out_occupant: Union["Occupant", None] = None
                    
    def _load_summary_data(self):
        """Function to load the "exited" and "remaining" door entries, which have a slightly different data structure"""
        self.usage = TimeSeries(time=[_ for _ in self._data.keys()], values=[int(_) for _ in self._data.values()])
            
    def _load_individual_door_data(self):
        """Load all of the potential fields for an individual door"""
        self._load_safe_time_series(source_data=self._data, data_key="usage", target="usage", data_type=int)
        self._load_safe_time_series(source_data=self._data, data_key="width", target="width")
        self._load_safe_time_series(source_data=self._data, data_key="boundary", target="boundary")
        self._load_safe_time_series(source_data=self._data, data_key="queueingDoorUsage", target="queueing_usage")
        
        direction_identifiers: Optional[Union[re.Match, None]] = re.search(r'\[[-+][XYZ]\]', self.name)
        
        if direction_identifiers != None:
            identifier: str = direction_identifiers.group(0)
            if identifier == Direction.X_POSITIVE:
                self.direction = Direction.X_POSITIVE
            elif identifier == Direction.X_NEGATIVE:
                self.direction = Direction.X_NEGATIVE
            elif identifier == Direction.Y_POSITIVE:
                self.direction = Direction.Y_POSITIVE
            elif identifier == Direction.Y_NEGATIVE:
                self.direction = Direction.Y_NEGATIVE
            elif identifier == Direction.Z_POSITIVE:
                self.direction = Direction.Z_POSITIVE
            elif identifier == Direction.Z_NEGATIVE:
                self.direction = Direction.Z_NEGATIVE


from typing import Iterable, List, Union


class ADataCollection():
    """Base class for any collection of data"""
    
    def __init__(self, *elements: Union[Iterable, List]):
        if type(*elements) == list:
            self._elements = iter(*elements)
        else:
            self._elements = tuple(*elements)

    def __getitem__(self, index):
        return self._elements[index]

    def __iter__(self):
        return self._elements.__iter__()

    def __len__(self):
        return len(self._elements)

    def __contains__(self, value):
        return value in self._elements

    def __repr__(self):
        return "[" + ",\n".join(str(e) for e in self._elements) + "]"

    def clear_cache(self):
        """Remove all data from the internal cache that has been loaded so far to free memory.
        """
        for element in self._elements:
            element.clear_cache()


class Point():
    """Generic representation of a 3D Point.
    
    Args:
        x (float): X component of the point
        y (float): y component of the point
        z (float): z component of the point
        
    Attributes:
        x (float): X component of the point
        y (float): Y component of the point
        z (float): Z component of the point
    """
    
    def __init__(self, x: float = float('inf'), y: float = float('inf'), z: float = float('inf')):
        self.x: float = x
        self.y: float = y
        self.z: float = z
        
    def __str__(self) -> str:
        return f'Point(X={self.x}, Y={self.y}, Z={self.z})'


import distutils.command.build
from setuptools import setup


# Override build command
class BuildCommand(distutils.command.build.build):
    def initialize_options(self):
        distutils.command.build.build.initialize_options(self)
        self.build_base = '../build'

setup(
    cmdclass={"build": BuildCommand},
)


from typing import Iterable, List, Union
from theng.data.collection import ADataCollection
from theng.data.pathfinder.groups.group import Group


class GroupCollection(ADataCollection):
    """Collection of :class:`~theng.data.pathfinder.Group` objects that provides array-like interface to make accessing Group data a little easier.
    
    Example:
        Elements in the Collection can be accessed in the following ways:
            - GroupCollection[str] => Gets the first Group object whose name matches the provided string.
            - GroupCollection[int] => Returns the Group object at the index specified by int.

    Arguments:
        groups (Union[Iterable, List]): Group objects to put in the collection
    """
    
    def __init__(self, *groups: Union[Iterable, List]):
        super().__init__(*groups)
        
    def __getitem__(self,  key: Union[int, str]) -> Union[Group, List[Group]]:
        if type(key) == int:
            return self._elements[key]
        else:
            return next(group for group in self._elements if (group.name == key))


"""Tag support is preliminary at this point. This currently serves as a data module for Occupants. 
There is hope for more model-level Tag functionality in the future.
"""

from typing import Dict, Optional
from theng.data.a_data_object import ADataObject


class Tag(ADataObject):
    """Represents a single Tag applied to an object in a Pathfinder simulation

    Args:
        name (str): The name of the Tag
        data (Dict): Raw data dict for the tag.
        
    Attributes:
        usage (Optional[float]): The amount of time that this Tag was applied to the relevant Occupant
        last_added(Optional[float]): The last timestep when this Tag was added to the relevant Occupant
        last_removed(Optional[float]): The last timestep when this Tag was removed from the relevant Occupant
    """
    
    def __init__(self, name: str, data: Dict):
        super().__init__()
        self.name = name
        self._data = data
        
        self.usage: Optional[float] = None
        self.last_added: Optional[float] = None
        self.last_removed: Optional[float] = None
        
        self._load()
    
    def _load(self) -> None:
        self._load_safe(source_data=self._data, data_key="usageTime", target="usage")
        self._load_safe(source_data=self._data, data_key="lastAdded", target="last_added")
        self._load_safe(source_data=self._data, data_key="lastRemoved", target="last_removed")